!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Author: Lukas EXL, last changes: 07.10.2014 !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine sinc_logr(S,S0,delta,tau,omega,C0)
!
! logr approximation by exponential sums of the form
! 
! log(r) ~ C0  - sum_l w_l exp(tau_l * r^2), l = -S ... S
! r \in [ delta, 2 ],  1 >> delta > 0 (e.g. delta = 1e-3)
!

implicit none
integer(4), intent(in) :: S, S0
real(8), intent(in) :: delta
real(8), intent(inout) :: tau(S-S0+1), omega(S-S0+1), C0
integer(4) :: j, ns 
real(8), parameter :: c_0 = 1.d0 
real(8) :: x, h


ns = 2*S+1
h = c_0 * dlog(dble(S))/dble(S)


do j = S0,S
  x = dexp(sinh(dble(j)*h))
  if (dlog(x) <= -6.d0) then
    tau(j-S0+1) = x - x*x/2.d0 + x*x*x/3.d0 - x*x*x*x/4.d0 + x*x*x*x*x/5.d0 - x*x*x*x*x*x/6.d0
  else if (sinh(dble(j)*h) >= 6.d0) then
    tau(j-S0+1) = dlog(x) + 1.d0/x - 1.d0/2.d0/x/x + 1.d0/3.d0/x/x/x - 1.d0/4.d0/x/x/x/x + 1.d0/5.d0/x/x/x/x/x - 1.d0/6.d0/x/x/x/x/x/x
  else
    tau(j-S0+1) = dlog(1+x)
  endif
enddo

do j = S0,S
  omega(j-S0+1) = h*cosh(dble(j)*h)/(1.d0 + exp(-sinh(dble(j)*h)))
enddo

tau = tau/delta/delta
omega = omega/delta/delta

do j = S0,S
  if (sinh(dble(j)*h) <= -8.d0) then
    x = dexp(sinh(dble(j)*h))
    omega(j-S0+1) = h/2.d0 * cosh(dble(j)*h) * ( (1-x+x*x-x*x*x+x*x*x*x-x*x*x*x*x)*(1.d0 + x/2.d0 - x*x/12.d0 + x*x*x/24.d0 - 19.d0*x*x*x*x/720.d0) )
  else
    omega(j-S0+1) = omega(j-S0+1)/tau(j-S0+1)/2.d0 !h/2 * cosh(j*h)/(1.d0 + exp(-sinh(j*h))) / dlog(1.d0 + exp(sinh(j*h)))
  endif
enddo

 C0 = sum(omega*dexp(-tau)) 

end subroutine sinc_logr

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine getdim(S,S0,delta)
integer(4), intent(inout) :: S, S0
real(8), intent(in) :: delta
integer(4) :: j
real(8), parameter :: c_0 = 1.d0
real(8) :: tau, x, h

h = c_0 * dlog(dble(S))/dble(S)

loop_1: do j = -S,S
  x = dexp(sinh(dble(j)*h))
  if (dlog(x) <= -6.d0) then
    tau = x - x*x/2.d0 + x*x*x/3.d0 - x*x*x*x/4.d0 + x*x*x*x*x/5.d0 - x*x*x*x*x*x/6.d0
  else if (sinh(dble(j)*h) >= 6.d0) then
    tau = dlog(x) + 1.d0/x - 1.d0/2.d0/x/x + 1.d0/3.d0/x/x/x - 1.d0/4.d0/x/x/x/x + 1.d0/5.d0/x/x/x/x/x - 1.d0/6.d0/x/x/x/x/x/x
  else
    tau = dlog(1+x)
  endif

  if ( tau*2/delta/delta >= 2.22d-16 ) then 
    S0 = j
    exit loop_1
  endif
enddo loop_1

end subroutine getdim

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine sinc_logr_eval(N,nsinc_logr,wsinc_logr,C0, rho,abserr,relerr)
!
! output : absolute err: abserr; relative error: relerr

implicit none
integer(4), intent(in) :: N
real(8), intent(in) :: rho, C0
real(8), intent(in) :: wsinc_logr(N), nsinc_logr(N)
real(8), intent(inout) :: abserr,relerr
real(8) :: I

I = C0  - sum(wsinc_logr*dexp(-nsinc_logr*rho*rho)) !+ C1*rho*rho

abserr = abs(I-dlog(rho))
relerr = abserr*abs(dlog(rho))
end subroutine sinc_logr_eval


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine sinc_para_logr_test(N,nsinc_logr,wsinc_logr,C0,rho_min,rho_max,numb,abserr,relerr)
implicit none
integer(4), intent(in) :: numb, N
real(8), intent(in)  :: nsinc_logr(N),wsinc_logr(N), rho_min, rho_max, C0
real(8), intent(inout) :: abserr, relerr
integer(4) :: j
real(8) :: rho(numb),  aerr, rerr
	
rho = dexp( dlog(rho_min) + (dlog(rho_max) - dlog(rho_min))/(numb - 1d0) * (/ (j, j=0,numb-1) /) ) 
		
!print *, minval(rho), maxval(rho)

abserr = 0d0
relerr = 0d0
do j = 1,numb
    call sinc_logr_eval(N,nsinc_logr,wsinc_logr,C0, rho(j),aerr,rerr)
    !abserr = abserr + aerr/dble(numb) C2,
    abserr = max(abserr,aerr)
     relerr = max(relerr,rerr)
enddo

end subroutine sinc_para_logr_test  


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine sinc_control_logr(S,S0,delta,ep)
! ep-accuarte logr approximation of form c0  - sum_l w_l exp(-tau_l r²)
! in interval [delta,2], delta << 1, e.g. delta = 1e-3  
implicit none
real(8), intent(in) :: delta, ep
integer(4), parameter :: nmin = 150, nmax = 500
real(8), allocatable :: nsinc_logr(:), wsinc_logr(:)
integer(4), intent(inout) :: S, S0 
integer(4) :: j, printitn = 10
integer(4), parameter :: numb = 10000
real(8) ::  rho_min, rho_max = 2.d0, abserr, relerr, C0
	
rho_min = delta

loop_1: do j = nmin,nmax,5
   S = j
   call getdim(S,S0,delta)
  ! print *, S0, S
   allocate( wsinc_logr(S-S0+1), nsinc_logr(S-S0+1) )
   call sinc_logr(S,S0,delta/2.d0,nsinc_logr,wsinc_logr,C0)
   call sinc_para_logr_test(S-S0+1,nsinc_logr,wsinc_logr,C0,rho_min,rho_max,numb,abserr,relerr)
   deallocate(wsinc_logr, nsinc_logr) 
   if (modulo(j,printitn) == 0) then
	  print *, S-S0+1, abserr      
   end if
   if (abserr < ep) then 
	exit loop_1
   end if
enddo loop_1

!print *, '[',rho_min,',',rho_max,']'
!print *, 'ind: ', S-S0+1
!print *, '2S+1: ', 2*j+1
	  
end subroutine sinc_control_logr

!